{***********************************************************************************
 *                    WBar Component By Ken Friesen                                *
 *                         Creation Date Jan 9/97                                  *
 *                                                                                 *
 *  This component is a 16 bit 3-d Graphical bar chart designed to set weighted    *
 *  averages.  It supports optional wrapped labels and percentage labels.  Weights *
 *  can be selected and adjusted by either using the arrow keys or the mouse.      *
 *  Direct entry of weights can be done by right clicking the chart to display a   *
 *  dialog.  The maximum number of series is 12.  The CutToClipboard method        *
 *  copies a bitmap of the chart to the clipboard.  The depth divisor adjusts the  *
 *  depth of the 3-d effect.  Must compile with the resource file.                 *
 *                                                                                 *
 ***********************************************************************************

                                                                                 }

unit Wbar;

interface
uses SysUtils, WinTypes, WinProcs, Messages, Dialogs, Classes, Graphics, Controls,
     Forms, StdCtrls, ExtCtrls, Buttons, Mask, Clipbrd;

{Holds the data for the bars - will be used to create a dynamic array, sort of
 like a TList}

type
  TSeries = record
    Name: String;
    Weight: Integer;
    Color: TColor;
end;

const
MaxArrayElements = 12;    {Max No of series is 12}

SeriesColors: Array [1..MaxArrayElements] of TColor =
              (clRed, clLime, clAqua , clYellow, clFuchsia, $008000FF,
               clOlive, $00FF0080, $00C080FF, clTeal,$000080FF,$0080FF80);


{Custom Event types}
type TWeightChange = procedure(Sender: TObject; Series: Integer) of Object;
type TSelect = procedure(Sender: TObject; Series: Integer) of Object;

{The dynamic array for data storage}
type
  TData = array[1..12] of TSeries;
  PData = ^TData;

type TWBar = Class(TCustomControl)
private
  IncreaseBtn: TBitBtn;
  DecreaseBtn: TBitBtn;
  Data: PData;
  CopyClip: Boolean;
  KeyDelay,KeyRepeat: integer;
  FSeriesCount: Integer;
  FSelected: Integer;
  FBackgroundColor: TColor;
  FOnWeightChange: TWeightChange;
  FOnSelect: TSelect;
  FOnBtnDown, FOnBtnUp: TNotifyEvent;
  FDepthDivisor: Integer;
  FShowPercent: Boolean;
  FShowSeriesNames: Boolean;
  WBarCursor: TCursor;
  BtnDown: boolean;
  procedure paint; override;
  procedure PaintBars(Bar: TBitmap);
  procedure PaintLabels(Bar: TBitmap; ypos, SeriesHeight,x: Integer);
  function CalculatePoint(y:Integer): Integer;
  procedure FSetSeriesCount(Value: Integer);
  procedure FSetSelected(Series: Integer);
  procedure FSetDepthDivisor(Value: Integer);
  procedure FSetBackGroundColor(Color: TColor);
  procedure FDisplayPercent(Display: Boolean);
  procedure FDisplayNames(Display: Boolean);
  procedure ShowWBarDlg;

  procedure WMSize(var WMSize: TWMSize); Message WM_Size;
  procedure WMMouseMove(var WMMouse: TWMMouse); Message WM_MouseMove;
  procedure WMLButtonDown(var WMMouse: TWMMouse); Message WM_LButtonDown;
  procedure WMRButtonDown(var WMMouse: TWMMouse); Message WM_RButtonDown;
  procedure WMKeyDown(var Key: TWMKey); Message WM_KeyDown;
  procedure WMKeyUp(var Key: TWMKey);  Message WM_KeyUp;
  procedure WM_GetDlgCode(var code: TWMGetDlgCode); Message WM_GetDlgCode;
  procedure WMSetCursor(var ACursor: TWMSetCursor); Message WM_SetCursor;
  procedure IncreaseBtnMouseDown(Sender: TObject; Button: TMouseButton;
            Shift: TShiftState; X, Y: Integer);
  procedure IncreaseBtnMouseUp(Sender: TObject; Button: TMouseButton;
            Shift: TShiftState; X, Y: Integer);

  procedure delay(milliseconds: integer);
  function GetNextSeries(Direction: Boolean):Integer;

public
  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;
  procedure SetExactWeight(Series: Integer; Value: Integer);
  procedure SetSeriesWeight(Series: Integer; Value: Integer);
  procedure SetSeriesColor(Series: Integer; Color: TColor);
  procedure SetSeriesName(Series: Integer; Name: String);
  function GetSeriesWeight(Series: Integer): Integer;
  function GetSeriesColor(Series: Integer): TColor;
  function GetSeriesName(Series: Integer): String;
  function GetTop: Integer;
  procedure CopyToClipboard;

published
  property BackgroundColor: TColor Read FBackgroundColor Write FSetBackgroundColor;
  property SeriesCount: Integer Read FSeriesCount Write FSetSeriesCount;
  property Selected: Integer Read FSelected Write FSetSelected ;
  property DepthDivisor: Integer Read FDepthDivisor Write FSetDepthDivisor ;
  property ShowPercent: Boolean Read FShowPercent Write FDisplayPercent;
  property ShowSeriesNames: Boolean Read FShowSeriesNames Write FDisplayNames;
  property Cursor;
  property DragMode;
  property Enabled;
  property Font;
  property TabStop;
  property Tag;
  property ShowHint;
  property ParentFont;
  property Visible;
  property HelpContext;
  property OnClick;
  property OnDragDrop;
  property OnDragOver;
  property OnEndDrag;
  property OnEnter;
  property OnExit;
  property OnMouseDown;
  property OnMouseMove;
  property OnMouseUp;
  property OnKeyPress;
  property OnWeightChange: TWeightChange Read FOnWeightChange Write FOnWeightChange;
  property OnSelect: TSelect Read FOnSelect Write FOnSelect;
  property OnBtnDown: TNotifyEvent Read FOnBtnDown Write FOnBtnDown;
  property OnBtnUp: TNotifyEvent Read FOnBtnUp Write FOnBtnUp;
end;

type
  TWBarDialog = class(TComponent)
  private
    Form: TForm;
    OKBtn: TButton;
    CancelBtn: TButton;
    WBarChart: TWBar;
    Bevel: TBevel;
    Shape: TShape;
    SumLabel: TLabel;
    procedure OKBtnClick(Sender: TObject);
    procedure CancelBtnClick(Sender: TObject);
    procedure MaskEditExit(Sender: TObject);
    procedure MaskEditChange(Sender: TObject);
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure Show;
  end;



procedure Register;

implementation
{$R WBar.Res}

{************************ Constructor/Destructor Procedures *******************}

procedure Register;
begin
  RegisterComponents('IA', [TWBar]);
end;

constructor TWBar.Create(AOwner: TComponent);
begin
  Inherited Create(AOwner);
  CopyClip:= False;
  KeyDelay:=getprofileint('Windows','KeyboardDelay',2);
  KeyRepeat:= GetProfileInt('Windows','KeyboardSpeed',30);
  BackGroundColor:= clBtnFace;
  SeriesCount:= 3;
  DepthDivisor:= 10;  {This value is the number used to divide the width to get the 3d depth}
  ShowSeriesNames:= True;
  ShowPercent:= True;
  ControlStyle:= Self.ControlStyle+ [csOpaque];  {To avoid the flicker on repaint}
  Width:= 150;
  Height:= 250;
  TabStop:= True;
  IncreaseBtn:= TBitBtn.Create(Self);
  With IncreaseBtn do
  begin
    Parent:= Self;
    Top:=0;
    Left:=Width-15;
    Width:=15;
    Height:=15;
    Caption:='';
    Glyph.Handle:= LoadBitmap(hInstance,'WPLUS');
    Visible:= false;
    TabStop:=False;
    Tag:=10;
    ControlStyle:= ControlStyle+[csOpaque];  {To avoid the flicker on repaint}
    OnMouseDown:=IncreaseBtnMouseDown;
    OnMouseUp:= IncreaseBtnMouseUp;
  end;
  DecreaseBtn:= TBitBtn.Create(Self);
  With DecreaseBtn do
  begin
    Parent:= Self;
    Top:=0;
    Left:=0;
    Width:=15;
    Height:=15;
    Glyph.Handle:= LoadBitmap(hInstance,'WMINUS');
    TabStop:=False;
    Tag:=-10;
    Visible:= false;
    ControlStyle:= ControlStyle+[csOpaque]; {To avoid the flicker on repaint}
    OnMouseDown:=IncreaseBtnMouseDown;
    OnMouseUp:= IncreaseBtnMouseUp;
  end;
end;

destructor TWBar.Destroy;
begin
  IncreaseBtn.Free;
  DecreaseBtn.Free;
  FreeMem(Data, FSeriesCount*SizeOf(TSeries));
  inherited Destroy;
end;

{*************************** Drawing Procedures ***************************}

procedure TWBar.Paint;
var Bar,Backgrnd: TBitmap;
begin
  Bar:= TBitmap.Create;
  try
    Bar.Canvas.Font:= Font;
    Bar.Height := Height-(2*IncreaseBtn.Height);
    Bar.Width := Width;
    PaintBars(Bar);
    {Set the clipboard with the bitmap of the bar}
    If CopyClip then
    begin
      Clipboard.Clear;
      Clipboard.Assign(Bar);
      CopyClip:= False;
    end else begin
      {Paint the background where the buttons will be}
      With Canvas do
      begin
        Brush.Color:= FBackGroundColor;
        FillRect(Rect(0,0,Width,(2*IncreaseBtn.Height)));
      end;
      {Paint the bar}
      Canvas.CopyMode := cmSrcCopy;
      Canvas.Draw(0, (2*IncreaseBtn.Height), Bar);
    end;
  finally
    Bar.Destroy;
  end;
end;

{The nasty painting of the bars proc}
procedure TWBar.PaintBars(Bar: TBitmap);
var x,SeriesHeight,ypos,Depth: integer;
begin
  Depth:= Width div DepthDivisor;
  ypos:= Depth;
  With Bar.Canvas do
  begin
    Brush.Color:= FBackGroundColor;
    FillRect(Rect(0,0,Bar.Width,Bar.Height));
    for x:= SeriesCount downto 1 do
    begin
      if x>1 then
        SeriesHeight:= MulDiv(Data^[x].Weight,Bar.Height-Depth,1000)
      else SeriesHeight:= Bar.Height-ypos;
      If (Selected=x) and not (csDesigning in ComponentState) then {draw the selected selected series}
      begin
        Brush.Color:= clBlack;
        FrameRect(Rect(0,ypos,Width-Depth,ypos+SeriesHeight));
        Pen.Width:=4;
        Pen.Color:= clYellow;
        PolyLine([Point(pen.width-1,ypos+1),
                  Point(Width-Depth-pen.width+1,ypos+2),
                  Point(Width-Depth-pen.width+1,ypos+abs(SeriesHeight-pen.width)+1),
                  Point(pen.width-1,ypos+abs(SeriesHeight-pen.width)+1),
                  Point(pen.width-1,ypos+1)]);
        Pen.Color:= clBlack;
        Pen.Width:=1;
        Brush.Color:= Data^[x].Color;
        Rectangle(4,ypos+4,Width-Depth-4,ypos+Abs(SeriesHeight-4));
        Brush.Color:= Data^[x].Color;
      end else begin  {draw the series}
        Brush.Color:= Data^[x].Color;
        Rectangle(0,ypos,Width-Depth,ypos+SeriesHeight);
      end;
      {Draw the 3d part Note: Calculatepoint proc provides the adjustment for perspective}
      Pen.Width:=1;
      Polygon([Point(Width-Depth-Pen.Width,ypos),
               Point(Width-Pen.Width,CalculatePoint(ypos)),
               Point(Width-Pen.Width,CalculatePoint(ypos+SeriesHeight)),
               Point(Width-Depth-Pen.Width,ypos+SeriesHeight-Pen.Width)]);
      if x= SeriesCount then
        Polygon([Point(0,Depth),
                 Point(Width-Depth-Pen.Width,Depth),
                 Point(Width,0),
                 Point(Depth*2-Pen.Width,0)]);
      if FShowSeriesNames or FShowPercent then PaintLabels(Bar,ypos,SeriesHeight,x);
      ypos:= ypos+SeriesHeight;
    end;
  end;
end;

procedure TWBar.PaintLabels(Bar: TBitmap; ypos, SeriesHeight,x: Integer);
var i,z,Start,Space,TxtHeight: Integer;
    Str: String;
    StrList: TStringList;
begin
  Str:= Data^[x].Name;
  TxtHeight:= Bar.Canvas.TextHeight(Str);
  SetBKMode(Bar.Canvas.Handle,TRANSPARENT); {Labels should be transparent}
  StrList:= TStringList.Create;
  try
    if x=FSelected then z:= 13 else z:=5;  {if selected then there is less room for Labels - z must be bigger}
    Start:=1;
    Space:=1;
    if FShowPercent and (TxtHeight<SeriesHeight-z) then {Add the % label}
      StrList.Add(IntToStr(Data^[x].Weight div 10)+DecimalSeparator+IntToStr(Data^[x].Weight mod 10)+'%');
    if FShowSeriesNames then for i:= 0 to Length(Str) do
    begin  {Add the Series Caption-  do the word wrapping if there is room and it is required}
      if TxtHeight*(StrList.Count+1)<SeriesHeight-z then
      begin
        if (Str[i]=#32)  then Space:= i;
        if (Bar.Canvas.TextWidth(Copy(Str,Start,i-Start))>= Width-(Width div DepthDivisor)-16) then
        begin
          If StrList.Count>0 then StrList.Insert(StrList.Count-1,Copy(Str,Start,Space-Start))
          else StrList.Add(Copy(Str,Start,Space-Start));
          Start:= Space+1;
        end;
        if (i=Length(Str)) and (TxtHeight*(StrList.Count+1)<SeriesHeight-z) then
        begin
          if StrList.Count>0 then
            StrList.Insert(StrList.Count-1,Copy(Str, Start, Length(Str)-Start+1))
          else StrList.Add(Copy(Str, Start, Length(Str)-Start+1));
        end;
      end;
    end;
    {We now have our label string list - just center it and paint}
    For i:= 0 to StrList.Count-1 do
      With Bar.Canvas do
        TextOut(((Width-(Width div DepthDivisor)) div 2)-(TextWidth(StrList[i]) div 2),
        YPos+(SeriesHeight div 2)-(StrList.Count*TextHeight(Str) div 2)+(i*TextHeight(Str)), StrList[i]);
    i:= StrList.Count;
  finally StrList.Free;
  end;
end;

{Provides vertical back point in perspective given the front of the trapezoid}
{Assumes a vanishing point of Width points to the NE of the top of the bar}
function TWBar.CalculatePoint(y:Integer): Integer;
var Slope,Intercept: Real;
    AHeight,ADepth, AWidth,x: Integer;
begin
  ADepth:= Width div DepthDivisor;
  AHeight:= Height-ADepth;
  AWidth:= Width-ADepth;
  Slope:= (ADepth-Width-y)/(Width);
  Intercept:= y-(slope*AWidth);
  x:= Round(Slope*Width + Intercept);
  CalculatePoint:=x;
end;

{********************* Public Procedures & Functions **********************}

procedure TWBar.CopyToClipboard;
begin
  CopyClip:= True;
  Repaint;
end;

{Sets the weight of one series and adjusts the rest propotionately to make series sum = 1000}
procedure TWBar.SetSeriesWeight(Series: Integer; Value: Integer);
var x,TmpInt,Total: Integer;
begin
  try
    If (Series>FSeriesCount) then Raise Exception.Create('Series out of Range!');
    If Series<1 then Exit;
    If Value>1000 then begin Value:=1000; MessageBeep(mb_OK); end;
    If Value<0 then begin Value:= 0; MessageBeep(mb_OK); end;
    Total:= Value;
    {Set the other series proportionately}
    for x:= 1 to FSeriesCount do
    begin
      if x<> Series then
      begin
        If Data^[Series].Weight=1000 then TmpInt:= (Data^[Series].Weight-Value) div (FSeriesCount-1)
        else TmpInt:=MulDiv(Data^[Series].Weight-Value, Data^[x].Weight ,1000-Data^[Series].Weight);
        If (Data^[x].Weight+TmpInt>=0) and (Data^[x].Weight+TmpInt<=1000) then
          Data^[x].Weight:= Data^[x].Weight+TmpInt
        else if Data^[x].Weight+TmpInt<0 then
          Data^[x].Weight:=0
        else if Data^[x].Weight+TmpInt>1000 then
          Data^[x].Weight:=1000;
        Total:= Total+Data^[x].Weight;
      end;
    end;
    {If there is any rounding errors - get rid of them by randomly increasing series}
    While Total<>1000 do
      If Total<1000 then
      begin
        GetNextSeries(True);
        If Data^[x].Weight<1000 then
        begin
          Inc(Data^[x].Weight);
          Inc(Total);
        end;
      end else begin
        GetNextSeries(True); 
        If Data^[x].Weight<1000 then
        begin
          Dec(Data^[x].Weight);
          Dec(Total);
        end;
      end;
    Data^[Series].Weight:=Value;
    Repaint;
    If Assigned(FOnWeightChange) and not (csDesigning in ComponentState) then
    FOnWeightChange(Self,Series);
  except
    ShowMessage('Error: Series out of Range!'+#10#13+'Must be between 1 & '+IntToStr(FSeriesCount));
  end;
end;


{No check to confirm series sum=1000, use with caution, - always setting all
series so that sum=1000.  Used to set multiple series at once}
procedure TWBar.SetExactWeight(Series: Integer; Value: Integer);
var Total, x: integer;
begin
  Data^[Series].Weight:= Value;
  Total:=0;
  for x:= 1 to FSeriesCount do Total:= Total+Data^[x].Weight;
  {Execute the OnWeightChange event - only if the total is correctly summed}
  If Assigned(FOnWeightChange) and not (csDesigning in ComponentState) and (Total=1000) then
    FOnWeightChange(Self,Series);
  Repaint;
end;


procedure TWBar.SetSeriesColor(Series: Integer; Color: TColor);
begin
  Data^[Series].Color:= Color;
  Repaint;
end;

procedure TWBar.SetSeriesName(Series: Integer; Name: String);
begin
  Data^[Series].Name:=Name;
  Repaint;
end;

function TWBar.GetSeriesWeight(Series: Integer): Integer;
begin
   GetSeriesWeight:= Data^[Series].Weight;
end;


function TWBar.GetSeriesColor(Series: Integer): TColor;
begin
  GetSeriesColor:= Data^[Series].Color;
end;

function TWBar.GetSeriesName(Series: Integer): String;
begin
  GetSeriesName:= Data^[Series].Name;
end;

procedure TWBar.FSetBackGroundColor(Color: TColor);
begin
  FBackGroundColor:= Color;
  Repaint;
end;

{Lets the person know how many points to offset the top of the WBar to get the top
 of the painted bar}
function TWBar.GetTop: Integer;
begin
  GetTop:= (IncreaseBtn.Height*2);
end;

{********************* Property Setting Procedures ***************************}

procedure TWBar.FSetDepthDivisor(Value: Integer);
begin
  try
    If (Value<5)then
    Raise Exception.CreateFmt('Divisor (%d) is not a valid number', [Value]);
    FDepthDivisor:=Value;
    If IncreaseBtn<>nil then
    begin
      IncreaseBtn.Left:= Width-IncreaseBtn.Width;
      DecreaseBtn.Left:= (Width div DepthDivisor)*2;
    end;
    FSetBackGroundColor(FBackGroundColor);
  except
    Showmessage(Format('Divisor (%d) is not a valid number', [Value])+
                #10#13+#10#13+ 'Must be above 4');
  end;
end;

procedure TWBar.FSetSelected(Series: Integer);
begin
  try
    If (Series<0) or (Series > FSeriesCount) then
    Raise Exception.CreateFmt('Series (%d) is not a valid number', [Series]);
  If (not Enabled) or (Series=0) then
  begin
    IncreaseBtn.Visible:=False;
    DecreaseBtn.Visible:=False;
  end else begin
    IncreaseBtn.Visible:=True;
    DecreaseBtn.Visible:=True;
  end;
  FSelected:=Series;
  FSetBackGroundColor(FBackGroundColor);
  If Assigned(FOnSelect) and not (csDesigning in ComponentState) then
    FOnSelect(Self,Series);
  Repaint;
  except
    Showmessage(Format('Series "%d" is not a valid number', [Series])+
                #10#13+#10#13+
                Format('Must be between 1 and %d, inclusive.', [FSeriesCount]));
  end;
end;

procedure TWBar.FSetSeriesCount(Value: Integer);
var x,Total: integer;
begin
  try
    If (Value > MaxArrayElements) or (Value<2) then
      Raise Exception.CreateFmt('Seriescount %d is not a valid number', [Value]);
    If Value<> FSeriesCount then
      begin
        {Reset all data}
        if FSeriesCount>0 then FreeMem(Data, FSeriesCount*SizeOf(TSeries));
        Data:= AllocMem(Value*Sizeof(TSeries));
        Total:=0;
        for x:= 1 to Value do
        With Data^[x] do
        begin
          Color:= SeriesColors[x];
          Weight:= 1000 div Value;
          Name:='Series '+IntToStr(x);
          Total:= Total+ Weight;
        end;
        FSeriesCount:= Value;
        {Adjust for rounding errors}
        While Total<1000 do
        begin
          x:= GetNextSeries(True);
          Inc(Data^[x].Weight);
          Inc(Total);
        end;
        If Assigned(FOnWeightChange) and not (csDesigning in ComponentState) then
          FOnWeightChange(Self,0);
        Repaint;
      end;
  except
    Showmessage(Format('Seriescount %d is not a valid number', [Value])+
                #10#13+#10#13+
                Format('Must be between 2 and %d, inclusive.', [MaxArrayElements]));
  end;
end;

procedure TWBar.FDisplayPercent(Display: Boolean);
begin
  FShowPercent:= Display;
  Repaint;
end;

procedure TWBar.FDisplayNames(Display: Boolean);
begin
  FShowSeriesNames:= Display;
  Repaint;
end;

{***************************** Button Procedures ****************************}

procedure TWBar.IncreaseBtnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  BtnDown:= True;
  SetSeriesWeight(FSelected, GetSeriesWeight(FSelected)+ (Sender as TBitBtn).Tag);
  Delay(1000 div KeyDelay);
  {If button is still down - keep executing until button comes up - like a keyboard key}
  While BtnDown do
  begin
    Delay(1);
    SetSeriesWeight(FSelected, GetSeriesWeight(FSelected)+ (Sender as TBitBtn).Tag);
  end;
  If Assigned(FOnBtnDown) and not (csDesigning in ComponentState) then
    FOnBtnDown(Sender);
end;

procedure TWBar.IncreaseBtnMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  If Assigned(FOnBtnUp) and not (csDesigning in ComponentState) then
    FOnBtnUp(Sender);
  BtnDown:=False;
  SetFocus;
end;

{***************************** Utility Functions ******************************}

procedure TWBar.delay(milliseconds: integer);
var x: longint;
begin
  x:= GetTickCount;
  while (GetTickCount-x)< milliseconds do Application.ProcessMessages;
end;

function TWBar.GetNextSeries(Direction: Boolean):Integer;
var TakeRemainder: integer;
begin
  Randomize;
  TakeRemainder:= Random(FSeriesCount)+1;
  If Direction then
  begin
    Repeat
      if TakeRemainder= FSeriesCount then TakeRemainder:=1 else Inc(TakeRemainder);
    until TakeRemainder<>FSelected;
  end else begin
    Repeat
      if TakeRemainder= 1 then TakeRemainder:=FSeriesCount else Dec(TakeRemainder);
    until TakeRemainder<>FSelected;
  end;
  GetNextSeries:= TakeRemainder;
end;


{***************************** Message Handlers *******************************}

{Position the buttons}
procedure TWBar.WMSize(var WMSize: TWMSize);
begin
  Inherited;
    IncreaseBtn.Left:= WMSize.Width-IncreaseBtn.Width;
    DecreaseBtn.Left:= (WMSize.Width div DepthDivisor)*2;
    FSetBackGroundColor(FBackGroundColor);
end;

{Set the cursor}
procedure TWBar.WMMouseMove(var WMMouse: TWMMouse);
var x: integer;
begin
  inherited;
  If not (csDesigning in ComponentState) then
  begin
    x:= Width div DepthDivisor;
    If (WMMouse.xpos < (Width - x)) and (WMMouse.ypos > (x+(2*IncreaseBtn.Height))) then
       cursor:= WBarCursor else cursor:= crDefault;
  end;
end;

{Select the correct series}
procedure TWBar.WMLButtonDown(var WMMouse: TWMMouse);
var x,ypos, Depth: integer;
begin
  inherited;
  If not (csDesigning in ComponentState) then
  begin
    Depth:= Width div DepthDivisor;
    ypos:= Depth+(2*IncreaseBtn.Height);
    If (WMMouse.xpos < (Width - Depth)) and (WMMouse.ypos > ypos) then
    begin
      for x:= FSeriesCount downto 1 do
      begin
        ypos:= ypos+ MulDiv(Data^[x].Weight,Height-Depth-(2*IncreaseBtn.Height),1000);
        if WMMouse.ypos<= ypos then
        begin
          if FSelected=x then Selected:= 0 else Selected:= x;
          x:=1;
          Repaint;
          SetFocus;
        end;
      end;
    end;
  end;
end;

{Open the WBar Dialog}
procedure TWBar.WMRButtonDown(var WMMouse: TWMMouse);
var x,ypos, Depth: integer;
begin
  inherited;
  if not (csDesigning in ComponentState) then
  begin
    Depth:= Width div DepthDivisor;
    ypos:= Depth+(2*IncreaseBtn.Height);
    If (WMMouse.xpos < (Width - Depth)) and (WMMouse.ypos > ypos) then
    begin
      SetFocus;
      ShowWBarDlg;
    end;
  end;
end;

{Allow Component to capture the Arrow key key presses}
procedure TWBar.WM_GetDlgCode(var Code: TWMGetDlgCode);
begin
  if not (csDesigning in ComponentState) then
     Code.Result:= DLGC_WantArrows;
end;

{Execute the Arrow key Commands}
procedure TWBar.WMKeyDown(var Key: TWMKey);
begin

  Case Key.CharCode of
  VK_UP:   begin
             If FSelected=SeriesCount then Selected:= 1 else
             Selected:= FSelected+1;
           end;
  VK_Down: begin
             If (FSelected=0) or (FSelected=1) then Selected:= SeriesCount
             else Selected:= FSelected-1;
           end;

  VK_Right: If FSelected<>0 then
            begin
              SetSeriesWeight(FSelected, GetSeriesWeight(FSelected)+ 10);
              If Assigned(FOnBtnDown) and not (csDesigning in ComponentState) then
              FOnBtnDown(Self);
            end;
  VK_Left: If FSelected<>0 then
           begin
              SetSeriesWeight(FSelected, GetSeriesWeight(FSelected)- 10);
              If Assigned(FOnBtnDown) and not (csDesigning in ComponentState) then
              FOnBtnDown(Self);
            end;
  end;
  inherited;
end;

{Stop the arrow key commands}
procedure TWBar.WMKeyUp(var Key: TWMKey);
begin

  Case Key.CharCode of

  VK_Right: If FSelected<>0 then
              If Assigned(FOnBtnUp) and not (csDesigning in ComponentState) then
              FOnBtnUp(Self);

  VK_Left: If FSelected<>0 then
             If Assigned(FOnBtnUp) and not (csDesigning in ComponentState) then
             FOnBtnUp(Self);
  end;
  inherited;
end;

procedure TWBar.WMSetCursor(var ACursor: TWMSetCursor);
begin
  inherited;
  if WBarCursor= 0 then WBarCursor:= Cursor;
end;

{********************* Dialog Section ***********************************}


procedure TWBar.ShowWBarDlg;
var WBarDialog: TWBarDialog;
begin
  WBarDialog:= TWBarDialog.Create(Self);
  WBarDialog.Show;
  WBarDialog.Free;
end;

{Create the dialog and set its values}
constructor TWBarDialog.Create(AOwner : TComponent);
var
   aHandle : THandle;
   Ypos,x : integer;
   ALabel: TLabel;
   AMaskEdit: TMaskEdit;
begin
  inherited Create(AOwner);
  WBarChart:= (AOwner as TWBar);
  Form:= TForm.Create(application);
  with form do begin
    Width:= 400;
    Height:= 135 + (22*(WBarChart.SeriesCount-2));
    Font.Name:= 'MS Sans Serif';
    Font.Size:= 9;
     BorderStyle:= bsDialog;
     Position:= poScreenCenter;
     Caption:= 'Set Series Weights';
     scaled:=False;
  end;
  OKBtn:= TButton.Create(Form);
  With OKBtn do
  begin
    Caption:= '&OK';
    Top:= 10;
    Left:= 330;
    Width:= 52;
    Height:= 25;
    Parent:= Form;
    OnClick:= OKBtnClick;
  end;
  CancelBtn:= TButton.Create(Form);
  With CancelBtn do
  begin
    Parent:= Form;
    Caption:= '&Cancel';
    Top:= 43;
    Left:= 330;
    Width:= 52;
    Height:= 25;
    OnClick:= CancelBtnClick;
  end;
  Bevel:= TBevel.Create(Form);
  With Bevel do
  begin
    Parent:= Form;
    Top:= 10;
    Left:= 10;
    Width:= 310;
    Height:= Form.ClientHeight-20;
    Shape:= bsFrame;
  end;

  Shape:= TShape.Create(Form);
  With Shape do
  begin
    Parent:= Form;
    Left:= 260;
    Top:= 22+(22*WBarChart.SeriesCount);
    Width:= 50;
    Height:=2;
  end;

  SumLabel:= TLabel.Create(Form);
  With SumLabel do
  begin
    Parent:= Form;
    Left:= 264;
    Top:= 30+(22*WBarChart.SeriesCount);
    Width:= 50;
    Height:=16;
    AutoSize:= False;
    Caption:='100'+DecimalSeparator+'0%'
  end;
  For x:= WBarChart.SeriesCount-1 downto 0 do
  begin
    AMaskEdit:= TMaskEdit.Create(Form);
    With AMaskEdit do
    begin
      Parent:= Form;
      Name:='EditMask' + IntToStr(x+1);
      Top:= 20+(22*(WBarChart.SeriesCount-x-1));
      Left:= 260;
      Width:= 50;
      Cursor:= crIBeam;
      AutoSize:= False;
      EditText:= IntToStr(WBarChart.GetSeriesWeight(x+1));
      If StrToInt(EditText)<10 then EditText:='0'+EditText;
      EditMask:= '!999'+DecimalSeparator+'9%;0; ';
      OnExit:= MaskEditExit;
      OnChange:= MaskEditChange;
    end;
    ALabel:= TLabel.Create(Form);
    With ALabel do
    begin
      Parent:= Form;
      Top:= 22+(22*(WBarChart.SeriesCount-x-1));
      Left:= 16;
      Width:= 230;
      AutoSize:= False;
      Caption:= WBarChart.GetSeriesName(x+1);
    end;
  end;
  MaskEditChange(AMaskEdit);
end;

destructor TWBarDialog.Destroy;
var x: integer;
begin  
   Form.Free;
   inherited Destroy;
end;

procedure TWBarDialog.Show;
begin
   Form.ShowModal;
end;


procedure TWBarDialog.OKBtnClick(Sender: TObject);
var Total,x: integer;
begin
  Total:=0;
  for x:= 1 to WBarChart.SeriesCount do
     Total:= Total+ StrToInt(TMaskEdit(Form.FindComponent('EditMask'+IntToStr(x))).Text);
  If Total<>1000 then MessageBox(Form.Handle,'Error: Weights must sum to 100%','Weight Bar',mb_OK)
  else begin
    for x:= 1 To WBarChart.SeriesCount do
    WBarChart.SetExactWeight(x,StrToInt(TMaskEdit(Form.FindComponent('EditMask'+IntToStr(x))).Text));
    WBarChart.Repaint;
    If Assigned(WBarChart.FOnWeightChange) then WBarChart.FOnWeightChange(WBarChart,0);
    Form.Close;
  end;
end;

procedure TWBarDialog.CancelBtnClick(Sender: TObject);
begin
  Form.Close;
end;

procedure TWBarDialog.MaskEditExit(Sender: TObject);
var Str: String;
begin
  Str:= (Sender as TMaskEdit).Text;
  (Sender as TMaskEdit).Text:= IntToStr(StrToIntDef(Str,0));
  If ((Sender as TMaskEdit).Text='') or ((Sender as TMaskEdit).Text='0') then
     (Sender as TMaskEdit).Text:='00'
  else If StrToInt((Sender as TMaskEdit).Text)>1000 then
  (Sender as TMaskEdit).Text:='1000';

end;

procedure TWBarDialog.MaskEditChange(Sender: TObject);
var x, Total: Integer;
    Str: String;
begin
  Str:= (Sender as TMaskEdit).Text;
  If Str='' then Exit;
  for x:= 0 to Length(Str) do
    if Str[x]=' ' then  Str[x]:='0';
  (Sender as TMaskEdit).Text:=Str;
  Total:=0;
  for x:= 1 to WBarChart.SeriesCount do
  begin
    Str:= TMaskEdit(Form.FindComponent('EditMask'+IntToStr(x))).Text;
    Total:= Total+ StrToInt(Str);
  end;
  SumLabel.Caption:= IntToStr(Total div 10) + DecimalSeparator + IntToStr(Total mod 10)+'%';
end;

end.
